import { BaseEvent } from '@/application/base/baseEvent'
import { SocketEvent } from '@/application/enums'
import type { MapEventTileWithTeleport } from '@/application/types'
import MapManager from '@/managers/mapManager'
import MapCharacter from '@/models/mapCharacter'
import MapEventTileRepository from '@/repositories/mapEventTileRepository'
import CharacterService from '@/services/characterMoveService'
import TeleportService from '@/services/characterTeleportService'

export default class CharacterMove extends BaseEvent {
  private readonly characterService = CharacterService
  private readonly STEP_DELAY = 100
  private readonly THROTTLE_DELAY = 230
  private readonly MAX_REQUEST_DISTANCE = 30 // Maximum allowed distance for movement requests
  private movementTimeout: NodeJS.Timeout | null = null
  private lastKnownPosition: { x: number; y: number } | null = null

  public listen(): void {
    this.socket.on(SocketEvent.MAP_CHARACTER_MOVE, this.handleEvent.bind(this))
  }

  private async handleEvent({ positionX, positionY }: { positionX: number; positionY: number }): Promise<void> {
    try {
      const mapCharacter = MapManager.getCharacterById(this.socket.characterId!)
      if (!mapCharacter?.getCharacter()) {
        this.logger.error('map:character:move error: Character not found or not initialized')
        return
      }

      const character = mapCharacter.getCharacter()
      const currentX = character.getPositionX()
      const currentY = character.getPositionY()

      // Enhanced throttling with position tracking
      const throttleKey = `movement_${this.socket.characterId}`
      if (this.isThrottled(throttleKey, this.THROTTLE_DELAY)) {
        return
      }

      // Validate current position against last known position
      const movementValidation = this.characterService.validateMovementDistance(currentX, currentY, this.lastKnownPosition, this.STEP_DELAY, mapCharacter.isMoving)

      if (!movementValidation.isValid) {
        this.logger.warn(`Suspicious movement detected: ${this.socket.characterId}`)
        // Force position reset
        character.setPositionX(this.lastKnownPosition!.x).setPositionY(this.lastKnownPosition!.y)
        this.broadcastMovement(character, false)
        return
      }

      // Validate requested position distance
      const requestValidation = this.characterService.validateRequestDistance(currentX, currentY, positionX, positionY, this.MAX_REQUEST_DISTANCE)

      if (!requestValidation.isValid) {
        this.logger.warn(`Invalid movement distance detected: ${this.socket.characterId}`)
        return
      }

      // If character is already moving to the same target position, ignore the request
      if (mapCharacter.isMoving && mapCharacter.currentPath?.length) {
        const lastPathPoint = mapCharacter.currentPath[mapCharacter.currentPath.length - 1]
        if (lastPathPoint && Math.abs(lastPathPoint.positionX - positionX) < 1 && Math.abs(lastPathPoint.positionY - positionY) < 1) {
          return
        }
      }

      // Cancel any ongoing movement
      this.cancelCurrentMovement(mapCharacter)

      // Update last known position
      this.lastKnownPosition = { x: currentX, y: currentY }

      // Calculate path to target position
      const path = await this.characterService.calculatePath(mapCharacter.character, Math.floor(positionX), Math.floor(positionY))

      if (!path?.length) {
        this.io.in(mapCharacter.character.map.id).emit(SocketEvent.MAP_CHARACTER_MOVEERROR, 'No valid path found')
        return
      }

      // Start new movement
      mapCharacter.isMoving = true
      mapCharacter.currentPath = path
      await this.moveAlongPath(mapCharacter)
    } catch (error: any) {
      this.logger.error('map:character:move error: ' + error.message)
    }
  }

  private cancelCurrentMovement(mapCharacter: MapCharacter): void {
    if (!mapCharacter.isMoving) return
    mapCharacter.isMoving = false
    mapCharacter.currentPath = null
  }

  private async moveAlongPath(mapCharacter: MapCharacter): Promise<void> {
    const character = mapCharacter.getCharacter()
    const path = mapCharacter.currentPath

    if (!path?.length) return

    let lastMoveTime = Date.now()
    let currentTile, nextTile

    try {
      for (let i = 0; i < path.length - 1; i++) {
        if (!mapCharacter.isMoving || mapCharacter.currentPath !== path) {
          return
        }

        // Ensure minimum time between moves using a single Date.now() call
        const timeSinceLastMove = Date.now() - lastMoveTime
        if (timeSinceLastMove < this.STEP_DELAY) {
          await new Promise((resolve) => setTimeout(resolve, this.STEP_DELAY - timeSinceLastMove))
        }

        currentTile = path[i]
        nextTile = path[i + 1]

        if (!currentTile || !nextTile || !this.isValidStep(currentTile, nextTile)) {
          this.logger.error('Invalid movement step detected')
          break
        }

        // Update character rotation and position in a single operation
        character
          .setRotation(CharacterService.calculateRotation(currentTile.positionX, currentTile.positionY, nextTile.positionX, nextTile.positionY))
          .setPositionX(nextTile.positionX)
          .setPositionY(nextTile.positionY)

        // Check for map events at the next tile
        const mapEventTile = await this.checkMapEvents(character, nextTile)
        if (mapEventTile) {
          if (mapEventTile.type === 'BLOCK') break
          if (mapEventTile.type === 'TELEPORT' && mapEventTile.teleport) {
            await this.handleTeleportMapEventTile(mapEventTile as MapEventTileWithTeleport)
            return
          }
        }

        // Broadcast movement
        this.broadcastMovement(character, true)

        // Apply movement delay between steps
        if (i < path.length - 2) {
          await new Promise((resolve) => setTimeout(resolve, this.STEP_DELAY))
        }

        // Update last known position and move time
        this.lastKnownPosition = {
          x: nextTile.positionX,
          y: nextTile.positionY
        }
        lastMoveTime = Date.now()
      }
    } finally {
      if (mapCharacter.isMoving && mapCharacter.currentPath === path) {
        this.finalizeMovement(mapCharacter)
      }
    }
  }

  private isValidStep(current: { positionX: number; positionY: number }, next: { positionX: number; positionY: number }): boolean {
    return Math.abs(next.positionX - current.positionX) <= 1 && Math.abs(next.positionY - current.positionY) <= 1
  }

  private async checkMapEvents(character: any, nextTile: { positionX: number; positionY: number }) {
    const mapEventTileRepository = new MapEventTileRepository()
    return mapEventTileRepository.getEventTileByMapIdAndPosition(character.getMap().getId(), Math.floor(nextTile.positionX), Math.floor(nextTile.positionY))
  }

  private async handleTeleportMapEventTile(mapEventTile: MapEventTileWithTeleport): Promise<void> {
    const teleport = mapEventTile.getTeleport()
    if (teleport) {
      await TeleportService.teleportCharacter(this.socket.characterId!, {
        targetMapId: teleport.getToMap().getId(),
        targetX: teleport.getToPositionX(),
        targetY: teleport.getToPositionY(),
        rotation: teleport.getToRotation()
      })
    }
  }

  private broadcastMovement(character: any, isMoving: boolean): void {
    this.io.in(character.map.id).emit(SocketEvent.MAP_CHARACTER_MOVE, {
      characterId: character.id,
      positionX: character.getPositionX(),
      positionY: character.getPositionY(),
      rotation: character.getRotation(),
      isMoving
    })
  }

  private finalizeMovement(mapCharacter: MapCharacter): void {
    // Clear any existing timeout
    if (this.movementTimeout) {
      clearTimeout(this.movementTimeout)
    }

    // Clear the current path immediately
    mapCharacter.currentPath = null

    // Set new timeout for movement state cleanup
    this.movementTimeout = setTimeout(() => {
      // Ensure the character is still in a valid state
      if (!mapCharacter.isMoving || !mapCharacter.currentPath) {
        mapCharacter.isMoving = false
        // Save the final position and broadcast it
        mapCharacter.savePosition().then(() => {
          this.broadcastMovement(mapCharacter.character, false)
        })
      }
    }, this.STEP_DELAY * 2) // Increased delay to ensure all movement processing is complete
  }
}
